#include <math.h>
#include "WritePDF.h"
#define MATHLIB_STANDALONE
#include <Rmath.h>

WritePDF::WritePDF() {}

void WritePDF::DrawQQplot(Vector & pvalueAll,Vector & pvalue1,Vector & pvalue5,String title)
{
   //Setup pvalue
   int N = pvalueAll.Length();
   int N1 = pvalue1.Length();
   int N5 = pvalue5.Length();
   int col = N+N1+N5;
   Matrix y;
   y.Dimension(5,col,_NAN_);
   pvalueAll.Sort();
   pvalue1.Sort();
   pvalue5.Sort();
int step = N/100;

   for(int i=0;i<N;i++)
   {
      y[0][i] = -log10((i*1.0+0.5)/N);
      if(i%step==0)
	 y[1][i] = y[0][i];
      y[2][i] = -log10(pvalueAll[i]);
   }
   for(int i=N;i<N+N1;i++)
   {
      y[0][i] = -log10((i*1.0-N+0.5)/N1);
      y[3][i] = -log10(pvalue1[i-N]);
   }
   for(int i=N+N1;i<N+N1+N5;i++)
   {
      y[0][i] = -log10((i*1.0-N-N1+0.5)/N5);
      y[4][i] = -log10(pvalue5[i-N-N1]);
   }
   // Setup Q-Q plot
   PDFLineChart chart;
   PDF pdf;

   String filename;
   filename.printf(title + ".qq.pdf");
   pdf.OpenFile(filename);

   chart.Reset();

   chart.title.printf(title + " QQ plot");
   chart.xAxis.label.printf("expected -log10(pvalue)");
   chart.yAxis.label.printf("observed -log10(pvalue)");

   chart.xAxis.SetMin(0.0);
   chart.xAxis.SetMax(-log10(pvalueAll.Min()));
   chart.yAxis.SetMin(0.0);
   chart.yAxis.SetMax(-log10(pvalueAll.Min()));

   chart.Dimension(4, N);
   chart.ShowMarker(0,false);
   chart.ShowMarker(1, true);
   chart.ShowMarker(2, true);
   chart.ShowMarker(3, true);
   chart.ShowLine(0, true);
   chart.ShowLine(1, false);
   chart.ShowLine(2, false);
   chart.ShowLine(3, false);
   //PDFLineStyle style[3]={ lsSolid, lsDashed, lsDotted };
   //chart.SetLineStyle(0,style[0]);
   // Draw Q-Q plot
   chart.SetSeriesLabel(0,"reference");
   chart.SetSeriesLabel(1,"all vars");
   chart.SetSeriesLabel(2,"maf<0.01");
   chart.SetSeriesLabel(3,"maf<0.05");

   chart.SetLineWeight(0,0.01);
   chart.SetMarkerShape(1,lmCircle);
   chart.SetMarkerShape(2,lmSquare);
   chart.SetMarkerShape(3,lmCross);
   chart.SetMarkerColor(1,1.0,0.0,0.0);
   chart.SetMarkerColor(2,0.0,1.0,0.0);
   chart.SetMarkerColor(3,0.0,0.0,0.0);
   chart.SetMarkerRadius(1,1.2);
   chart.SetMarkerRadius(2,1.5);
   chart.SetMarkerRadius(3,1.2);
   pdf.page.SetFontSize(14);

   chart.SetDataValues(y);
   chart.Draw(pdf);
   pdf.CloseFile();
}

void WritePDF::DrawQQplot(Vector & pvalueAll, String title, String legend)
{
   //Setup pvalue
   int N = pvalueAll.Length();
   Matrix y;
   y.Dimension(3,N,_NAN_);
   pvalueAll.Sort(); 

   for(int i=0;i<N;i++)
   {
      y[0][i] = -log10((i*1.0+0.5)/N);
      y[1][i] = y[0][i];
      y[2][i] = -log10(pvalueAll[i]);
   }
   // Setup Q-Q plot
   PDFLineChartwithPolygon chart;
   PDF pdf;

   String filename;
   filename.printf(title + ".qq.pdf");
   pdf.OpenFile(filename);
   chart.Reset();


   chart.title.printf(title + " QQ plot");
   chart.xAxis.label.printf("expected -log10(pvalue)");
   chart.yAxis.label.printf("observed -log10(pvalue)");

   chart.xAxis.SetMin(0.0);
   chart.xAxis.SetMax(-log10(pvalueAll.Min()));
   chart.yAxis.SetMin(0.0);
   chart.yAxis.SetMax(-log10(pvalueAll.Min()));

   chart.Dimension(2, N);
   chart.ShowMarker(0,false);
   chart.ShowMarker(1, true);
   chart.ShowLine(0, true);
   chart.ShowLine(1, false);
   chart.SetSeriesLabel(0,"expected");
   chart.SetSeriesLabel(1,legend.c_str());

   chart.SetLineWeight(0,0.01);
   chart.SetMarkerShape(1,lmCircle);
   chart.SetMarkerColor(1,1.0,0.0,0.0);
   chart.SetMarkerRadius(1,1.5);
   pdf.page.SetFontSize(14);

   chart.SetDataValues(y);

   //fill between the CI
   int n = N<1000 ? N:1000;
   double  * poly_x = new double [2*n];
   double * poly_y = new double [2*n];
   //set values to poly_x and poly_y
   int step = N/n;
double conf = 0.05;
   for(int i=0;i<N;i++)
   {
      int mod = i%step;
      if(mod==0)
      {
	 int j = i/step;
	 poly_x[j] = poly_x[2*n-j-1] = -log10((i*1.0+0.5)/N); 
	 poly_y[j] = -log10(qbeta((1.0-conf/2),i+1,N-i,1,0));
	 poly_y[2*n-1-j] = -log10(qbeta(conf/2,i+1,N-i,1,0));
      }
   }
/*
//troubleshoot
printf("N is %d\n",N);
for(int i=0;i<2*n;i++)
{
printf("%g %g\n",poly_x[i],poly_y[i]);
}
*/
   //set fill color to be light gray
chart.poly_x = poly_x;
chart.poly_y = poly_y;
chart.poly_n = 2*n;
   chart.Draw(pdf);

   pdf.CloseFile();
   if(poly_x) delete [] poly_x;
   if(poly_y) delete [] poly_y;
}

/*
   void WritePDF::DrawManhattanPlot(Vector & pvalueAll,intArray & positions)
   {
//Setup pvalue
int N = pvalueAll.Length();
int N1 = pvalue1.Length();
int N5 = pvalue5.Length();
int col = N+N1+N5;
Matrix y;
y.Dimension(5,col,_NAN_);
pvalueAll.Sort();
pvalue1.Sort();
pvalue5.Sort();

for(int i=0;i<N;i++)
{
y[0][i] = -log10((i+1)*1.0/(N+1));
y[1][i] = y[0][i];
y[2][i] = -log10(pvalueAll[i]);
}
for(int i=N;i<N+N1;i++)
{
y[0][i] = -log10((i-N+1)*1.0/(N1+1));
y[3][i] = -log10(pvalue1[i-N]);
}
for(int i=N+N1;i<N+N1+N5;i++)
{
y[0][i] = -log10((i-N-N1+1)*1.0/(N5+1));
y[4][i] = -log10(pvalue5[i-N-N1]);
}
// Setup Q-Q plot
PDFLineChart chart;
PDF pdf;

 */
